{% ckan_extends %}

{% block content %}
  {{ super() }}

  {# Use Fanstatic to include our custom JavaScript module.
     A <script> tag for the module will be inserted in the right place at the
     bottom of the page. #}

  {# TODO #}
  {#% resource 'datastore_ts/datastore_ts_popover.js' %#}

  {# Apply our JavaScript module to an HTML element. The data-module attribute,
     which can be applied to any HTML element, tells CKAN to initialize an
     instance of the named JavaScript module for the element.
     The initialize() method of our module will be called with this HTML
     element as its this.el object. #}
  <button data-module="datastore_ts_popover"
          class="btn"
          href="#">
    <i class="icon-info-sign"></i>
  </button>
{% endblock %}